@@ -92,11 +92,11 @@ module Control.Foldl (
     , Control.Foldl.mapM_
     , sink
 
-    -- * Generic Folds
+    -- ** Generic Folds
     , genericLength
     , genericIndex
 
-    -- * Container folds
+    -- ** Container Folds
     , list
     , revList
     , nub
@@ -123,6 +123,7 @@ module Control.Foldl (
     , _Fold1
     , premap
     , premapM
+    , postmapM
     , prefilter
     , prefilterM
     , predropWhile
@@ -159,8 +160,7 @@ import Data.Foldable (Foldable)
 import Data.Functor.Identity (Identity, runIdentity)
 import Data.Functor.Contravariant (Contravariant(..))
 import Data.HashMap.Strict (HashMap)
-import Data.Map.Strict (Map, alter)
-import Data.Maybe (fromMaybe)
+import Data.Map.Strict (Map)
 import Data.Monoid hiding ((<>))
 import Data.Semigroupoid (Semigroupoid)
 import Data.Functor.Extend (Extend(..))
@@ -209,6 +209,7 @@ import qualified Data.Semigroupoid
 {- $setup
 
 >>> import qualified Control.Foldl as Foldl
+>>> import Data.Functor.Apply (Apply(..))
 
 >>> _2 f (x, y) = fmap (\i -> (x, i)) (f y)
 
@@ -218,7 +219,7 @@ import qualified Data.Semigroupoid
 >>>         in Control.Foldl.Optics.prism Just maybeEither
 >>> :}
 
->>> both f (x, y) = (,) <$> f x <*> f y
+>>> both f (x, y) = (,) <$> f x <.> f y
 
 -}
 
@@ -300,7 +301,7 @@ instance Monoid b => Monoid (Fold a b) where
     mempty = pure mempty
     {-# INLINE mempty #-}
 
-    mappend = liftA2 mappend
+    mappend = (<>)
     {-# INLINE mappend #-}
 
 instance Num b => Num (Fold a b) where
@@ -434,7 +435,7 @@ instance (Monoid b, Monad m) => Monoid (FoldM m a b) where
     mempty = pure mempty
     {-# INLINE mempty #-}
 
-    mappend = liftA2 mappend
+    mappend = (<>)
     {-# INLINE mappend #-}
 
 instance (Monad m, Num b) => Num (FoldM m a b) where
@@ -978,7 +979,7 @@ foldByKeyMap f = case f of
   Fold (step0 :: x -> a -> x) (ini0 :: x) (end0 :: x -> b) ->
     let
       step :: Map k x -> (k,a) -> Map k x
-      step mp (k,a) = Map.alter addToMap k mp where
+      step !mp (k,a) = Map.alter addToMap k mp where
         addToMap Nothing         = Just $ step0 ini0 a
         addToMap (Just existing) = Just $ step0 existing a
 
@@ -1227,6 +1228,19 @@ premapM f (FoldM step begin done) = FoldM step' begin done
     step' x a = f a >>= step x
 {-# INLINABLE premapM #-}
 
+{-| @(postmapM f folder)@ returns a new 'FoldM' where f is applied to the final value.
+
+> postmapM return = id
+>
+> postmapM (f >=> g) = postmapM g . postmapM f
+
+> postmapM k (pure r) = k r
+-}
+postmapM :: Monad m => (a -> m r) -> FoldM m x a -> FoldM m x r
+postmapM f (FoldM step begin done) = FoldM step begin done'
+  where done' x = done x >>= f
+{-# INLINABLE postmapM #-}
+
 {-| @(prefilter f folder)@ returns a new 'Fold' where the folder's input is used
   only when the input satisfies a predicate f
 
@@ -1336,7 +1350,7 @@ type Handler a b =
 >>> fold (handles traverse sum) [[1..5],[6..10]]
 55
 
->>> fold (handles (traverse.traverse) sum) [[Nothing, Just 2, Just 7],[Just 13, Nothing, Just 20]]
+>>> fold (handles (traverse . traverse) sum) [[Nothing, Just 2, Just 7],[Just 13, Nothing, Just 20]]
 42
 
 >>> fold (handles (filtered even) sum) [1..10]
@@ -1369,7 +1383,7 @@ handles k (Fold step begin done) = Fold step' begin done
 
 > Foldl.foldOver f folder xs == Foldl.fold folder (xs^..f)
 
-> Foldl.foldOver (folded.f) folder == Foldl.fold (handles f folder)
+> Foldl.foldOver (folded . f) folder == Foldl.fold (handles f folder)
 
 > Foldl.foldOver folded == Foldl.fold
 
@@ -1430,7 +1444,7 @@ handlesM k (FoldM step begin done) = FoldM step' begin done
 
 {- | @(foldOverM f folder xs)@ folds all values from a Lens, Traversal, Prism or Fold monadically with the given folder
 
-> Foldl.foldOverM (folded.f) folder == Foldl.foldM (handlesM f folder)
+> Foldl.foldOverM (folded . f) folder == Foldl.foldM (handlesM f folder)
 
 > Foldl.foldOverM folded == Foldl.foldM
 
@@ -1476,10 +1490,8 @@ projection function. Returns the folded result grouped as a map keyed by the
 group.
 
 -}
-groupBy :: Ord g => (a -> g) -> Fold a r -> Fold a (Map g r)
-groupBy grouper (Fold f i e) = Fold f' mempty (fmap e)
-  where
-    f' !m !a = alter (\o -> Just (f (fromMaybe i o) a)) (grouper a) m
+groupBy :: Ord k => (a -> k) -> Fold a b -> Fold a (Map k b)
+groupBy f = premap (\(!a) -> (f a, a)) . foldByKeyMap
 {-# INLINABLE groupBy #-}
 
 {-| Combine two folds into a fold over inputs for either of them.
